@@ -4,7 +4,7 @@
 #
 # Copyright (C) 2004 Robert Kaye
 # Copyright (C) 2006-2009, 2011-2012, 2014 Lukáš Lalinský
-# Copyright (C) 2008-2011, 2014, 2018-2022 Philipp Wolfer
+# Copyright (C) 2008-2011, 2014, 2018-2024 Philipp Wolfer
 # Copyright (C) 2009 Carlin Mangar
 # Copyright (C) 2009 david
 # Copyright (C) 2010 fatih
@@ -12,7 +12,7 @@
 # Copyright (C) 2012, 2014-2015 Wieland Hoffmann
 # Copyright (C) 2013 Ionuț Ciocîrlan
 # Copyright (C) 2013-2014 Sophist-UK
-# Copyright (C) 2013-2014, 2018-2021 Laurent Monin
+# Copyright (C) 2013-2014, 2018-2022 Laurent Monin
 # Copyright (C) 2014 Johannes Dewender
 # Copyright (C) 2016 Rahul Raturi
 # Copyright (C) 2016 barami
@@ -23,6 +23,8 @@
 # Copyright (C) 2020 Ray Bouchard
 # Copyright (C) 2021 Gabriel Ferreira
 # Copyright (C) 2021 Louis Sautier
+# Copyright (C) 2022 Kamil
+# Copyright (C) 2022 skelly37
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -38,13 +40,20 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
-
-import builtins
-from collections import namedtuple
+try:
+    from charset_normalizer import detect
+except ImportError:
+    try:
+        from chardet import detect
+    except ImportError:
+        detect = None
+from collections import (
+    defaultdict,
+    namedtuple,
+)
 from collections.abc import Mapping
 from itertools import chain
 import json
-from locale import strxfrm as _strxfrm
 import ntpath
 from operator import attrgetter
 import os
@@ -58,6 +67,7 @@ import unicodedata
 from dateutil.parser import parse
 
 from PyQt5 import QtCore
+from PyQt5.QtGui import QDesktopServices
 
 from picard import log
 from picard.const import (
@@ -88,25 +98,53 @@ WIN_MAX_NODE_LEN = 255
 WIN_LONGPATH_PREFIX = '\\\\?\\'
 
 
-class LockableObject(QtCore.QObject):
+class ReadWriteLockContext:
+    """Context for releasing a locked QReadWriteLock
+    """
+    def __init__(self):
+        self.__lock = QtCore.QReadWriteLock()
 
+    def lock_for_read(self):
+        self.__lock.lockForRead()
+        return self
+
+    def lock_for_write(self):
+        self.__lock.lockForWrite()
+        return self
+
+    def unlock(self):
+        self.__lock.unlock()
+
+    def __enter__(self):
+        pass
+
+    def __exit__(self, type, value, tb):
+        self.__lock.unlock()
+
+    def __bool__(self):
+        return self._entered > 0
+
+
+class LockableObject(QtCore.QObject):
     """Read/write lockable object."""
 
     def __init__(self):
         super().__init__()
-        self.__lock = QtCore.QReadWriteLock()
+        self.__context = ReadWriteLockContext()
 
     def lock_for_read(self):
         """Lock the object for read operations."""
-        self.__lock.lockForRead()
+        self.__context.lock_for_read()
+        return self.__context
 
     def lock_for_write(self):
         """Lock the object for write operations."""
-        self.__lock.lockForWrite()
+        self.__context.lock_for_write()
+        return self.__context
 
     def unlock(self):
         """Unlock the object."""
-        self.__lock.unlock()
+        self.__context.unlock()
 
 
 def process_events_iter(iterable, interval=0.1):
@@ -212,19 +250,20 @@ def system_supports_long_paths():
         system_supports_long_paths._supported = supported
         return supported
     except OSError:
-        log.info('Failed reading LongPathsEnabled from registry')
+        log.info("Failed reading LongPathsEnabled from registry")
         return False
 
 
-def normpath(path):
+def normpath(path, realpath=True):
     path = os.path.normpath(path)
-    try:
-        path = os.path.realpath(path)
-    except OSError as why:
-        # realpath can fail if path does not exist or is not accessible
-        # or on Windows if drives are mounted without mount manager
-        # (see https://tickets.metabrainz.org/browse/PICARD-2425).
-        log.warning('Failed getting realpath for "%s": %s', path, why)
+    if realpath:
+        try:
+            path = os.path.realpath(path)
+        except OSError as why:
+            # realpath can fail if path does not exist or is not accessible
+            # or on Windows if drives are mounted without mount manager
+            # (see https://tickets.metabrainz.org/browse/PICARD-2425).
+            log.warning("Failed getting realpath for `%s`: %s", path, why)
     # If the path is longer than 259 characters on Windows, prepend the \\?\
     # prefix. This enables access to long paths using the Windows API. See
     # https://docs.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation
@@ -234,8 +273,16 @@ def normpath(path):
 
 
 def win_prefix_longpath(path):
+    """
+    For paths longer then WIN_MAX_FILEPATH_LEN enable long path support by prefixing with WIN_LONGPATH_PREFIX.
+
+    See https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation
+    """
     if len(path) > WIN_MAX_FILEPATH_LEN and not path.startswith(WIN_LONGPATH_PREFIX):
-        path = WIN_LONGPATH_PREFIX + path
+        if path.startswith(r'\\'):  # UNC path
+            path = WIN_LONGPATH_PREFIX + 'UNC' + path[1:]
+        else:
+            path = WIN_LONGPATH_PREFIX + path
     return path
 
 
@@ -243,7 +290,15 @@ def is_absolute_path(path):
     """Similar to os.path.isabs, but properly detects Windows shares as absolute paths
     See https://bugs.python.org/issue22302
     """
-    return os.path.isabs(path) or (IS_WIN and os.path.normpath(path).startswith("\\\\"))
+    if IS_WIN:
+        # Two backslashes indicate a UNC path.
+        if path.startswith("\\\\"):
+            return True
+        # Consider a single slash at the start not relative. This is the default
+        # for `os.path.isabs` since Python 3.13.
+        elif path.startswith("\\") or path.startswith("/"):
+            return False
+    return os.path.isabs(path)
 
 
 def samepath(path1, path2):
@@ -279,31 +334,41 @@ def format_time(ms, display_zero=False):
 def sanitize_date(datestr):
     """Sanitize date format.
 
-    e.g.: "YYYY-00-00" -> "YYYY"
-          "YYYY-  -  " -> "YYYY"
+    e.g.: "1980-00-00" -> "1980"
+          "1980-  -  " -> "1980"
+          "1980-00-23" -> "1980-00-23"
           ...
     """
     date = []
-    for num in datestr.split("-"):
+    for num in reversed(datestr.split("-")):
         try:
             num = int(num.strip())
         except ValueError:
-            break
-        if num:
+            if num == '':
+                num = 0
+            else:
+                break
+        if num or (num == 0 and date):
             date.append(num)
+    date.reverse()
     return ("", "%04d", "%04d-%02d", "%04d-%02d-%02d")[len(date)] % tuple(date)
 
 
-_re_win32_incompat = re.compile(r'["*:<>?|]', re.UNICODE)
-def replace_win32_incompat(string, repl="_"):  # noqa: E302
+def replace_win32_incompat(string, repl="_", replacements=None):  # noqa: E302
     """Replace win32 filename incompatible characters from ``string`` by
        ``repl``."""
-    # Don't replace : with _ for windows drive
+    # Don't replace : for windows drive
     if IS_WIN and os.path.isabs(string):
-        drive, rest = ntpath.splitdrive(string)
-        return drive + _re_win32_incompat.sub(repl, rest)
+        drive, string = ntpath.splitdrive(string)
     else:
-        return _re_win32_incompat.sub(repl, string)
+        drive = ''
+
+    replacements = defaultdict(lambda: repl, replacements or {})
+    for char in {'"', '*', ':', '<', '>', '?', '|'}:
+        if char in string:
+            string = string.replace(char, replacements[char])
+
+    return drive + string
 
 
 _re_non_alphanum = re.compile(r'\W+', re.UNICODE)
@@ -349,7 +414,7 @@ def translate_from_sortname(name, sortname):
     """'Translate' the artist name by reversing the sortname."""
     for c in name:
         ctg = unicodedata.category(c)
-        if ctg[0] == "L" and unicodedata.name(c).find("LATIN") == -1:
+        if ctg[0] == "L" and unicodedata.name(c).find('LATIN') == -1:
             for separator in (" & ", "; ", " and ", " vs. ", " with ", " y "):
                 if separator in sortname:
                     parts = sortname.split(separator)
@@ -419,6 +484,14 @@ def run_executable(executable, *args, timeout=None):
     return ret.returncode, ret.stdout.decode(sys.stdout.encoding), ret.stderr.decode(sys.stderr.encoding)
 
 
+def open_local_path(path):
+    url = QtCore.QUrl.fromLocalFile(path)
+    if os.environ.get('SNAP'):
+        run_executable('xdg-open', url.toString())
+    else:
+        QDesktopServices.openUrl(url)
+
+
 _mbid_format = '[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'
 _re_mbid_val = re.compile(_mbid_format, re.IGNORECASE)
 def mbid_validate(string):  # noqa: E302
@@ -667,6 +740,17 @@ def album_artist_from_path(filename, album, artist):
     return album, artist
 
 
+def encoded_queryargs(queryargs):
+    """
+    Percent-encode all values from passed dictionary
+    Keys are left unmodified
+    """
+    return {
+        name: bytes(QtCore.QUrl.toPercentEncoding(str(value))).decode()
+        for name, value in queryargs.items()
+    }
+
+
 def build_qurl(host, port=80, path=None, queryargs=None):
     """
     Builds and returns a QUrl object from `host`, `port` and `path` and
@@ -679,11 +763,11 @@ def build_qurl(host, port=80, path=None, queryargs=None):
     url.setHost(host)
 
     if port == 443 or host in MUSICBRAINZ_SERVERS:
-        url.setScheme("https")
+        url.setScheme('https')
     elif port == 80:
-        url.setScheme("http")
+        url.setScheme('http')
     else:
-        url.setScheme("http")
+        url.setScheme('http')
         url.setPort(port)
 
     if path is not None:
@@ -744,14 +828,6 @@ def __convert_to_string(obj):
         return str(obj)
 
 
-def convert_to_string(obj):
-    log.warning("string_() and convert_to_string() are deprecated, do not use")
-    return __convert_to_string(obj)
-
-
-builtins.__dict__['string_'] = convert_to_string
-
-
 def load_json(data):
     """Deserializes a string or bytes like json response and converts
     it to a python object.
@@ -843,21 +919,22 @@ def find_best_match(candidates, no_match):
     return BestMatch(similarity=best_match.similarity, result=best_match)
 
 
-def get_qt_enum(cls, enum):
+def get_qt_enum(cls, attr_class):
     """
-    List all the names of attributes inside a Qt enum.
+    Generate all the names of attributes inside a class that are instances of a specific class
+
+    Args:
+        cls: the class in which to search attributes
+        attr_class: class of attributes to match
 
     Example:
         >>> from PyQt5.Qt import Qt
-        >>> print(get_qt_enum(Qt, Qt.CoordinateSystem))
+        >>> print(list(get_qt_enum(Qt, Qt.CoordinateSystem)))
         ['DeviceCoordinates', 'LogicalCoordinates']
     """
-    values = []
     for key in dir(cls):
-        value = getattr(cls, key)
-        if isinstance(value, enum):
-            values.append(key)
-    return values
+        if isinstance(getattr(cls, key), attr_class):
+            yield key
 
 
 def limited_join(a_list, limit, join_string='+', middle_string='…'):
@@ -896,6 +973,10 @@ def limited_join(a_list, limit, join_string='+', middle_string='…'):
     return join_string.join(start + [middle_string] + end)
 
 
+def countries_shortlist(countries):
+    return limited_join(countries, 6, '+', '…')
+
+
 def extract_year_from_date(dt):
     """ Extracts year from  passed in date either dict or string """
 
@@ -904,7 +985,7 @@ def extract_year_from_date(dt):
             return int(dt.get('year'))
         else:
             return parse(dt).year
-    except (TypeError, ValueError):
+    except (OverflowError, TypeError, ValueError):
         return None
 
 
@@ -1035,7 +1116,7 @@ def unique_numbered_title(default_title, existing_titles, fmt=None):
        based on given default title and existing titles
     """
     if fmt is None:
-        fmt = _(DEFAULT_NUMBERED_TITLE_FORMAT)
+        fmt = gettext_constants(DEFAULT_NUMBERED_TITLE_FORMAT)
 
     escaped_title = re.escape(default_title)
     reg_count = r'(\d+)'
@@ -1058,7 +1139,7 @@ def get_base_title_with_suffix(title, suffix, fmt=None):
        removing the suffix and number portion from the end.
     """
     if fmt is None:
-        fmt = _(DEFAULT_NUMBERED_TITLE_FORMAT)
+        fmt = gettext_constants(DEFAULT_NUMBERED_TITLE_FORMAT)
 
     escaped_suffix = re.escape(suffix)
     reg_title = r'(?P<title>.*?)(?:\s*' + escaped_suffix + ')?'
@@ -1073,7 +1154,7 @@ def get_base_title_with_suffix(title, suffix, fmt=None):
 def get_base_title(title):
     """Extract the base portion of a title, using the standard suffix.
     """
-    suffix = _(DEFAULT_COPY_TEXT)
+    suffix = gettext_constants(DEFAULT_COPY_TEXT)
     return get_base_title_with_suffix(title, suffix)
 
 
@@ -1092,19 +1173,94 @@ def any_exception_isinstance(error, type_):
     return any(isinstance(err, type_) for err in iter_exception_chain(error))
 
 
-def strxfrm(string):
-    """Transforms a string to one that can be used in locale-aware comparisons.
+ENCODING_BOMS = {
+    b'\xff\xfe\x00\x00': 'utf-32-le',
+    b'\x00\x00\xfe\xff': 'utf-32-be',
+    b'\xef\xbb\xbf': 'utf-8-sig',
+    b'\xff\xfe': 'utf-16-le',
+    b'\xfe\xff': 'utf-16-be',
+}
 
-    Wrapper around locale.strxfrm, that never throws OSError. If an OSError
-    occurs this function will return the string converted to lower case.
+
+def detect_file_encoding(path, max_bytes_to_read=1024*256):
+    """Attempts to guess the unicode encoding of a file based on the BOM, and
+    depending on avalibility, using a charset detection method.
+
+    Assumes UTF-8 by default if no other encoding is detected.
 
     Args:
-        string: The string to convert
+        path: The path to the file
+        max_bytes_to_read: Maximum bytes to read from the file during encoding
+        detection.
 
-    Returns: A string that can be compared locale-aware
+    Returns: The encoding as a string, e.g. "utf-16-le" or "utf-8"
     """
-    try:
-        return _strxfrm(string)
-    except OSError as err:
-        log.warning('strxfrm(%r) failed: %r', string, err)
-        return string.lower()
+    with open(path, 'rb') as f:
+        first_bytes = f.read(4)
+        for bom, encoding in ENCODING_BOMS.items():
+            if first_bytes.startswith(bom):
+                return encoding
+
+        if detect is None:
+            return 'utf-8'
+
+        f.seek(0)
+        result = detect(f.read(max_bytes_to_read))
+        if result['encoding'] is None:
+            log.warning("Couldn't detect encoding for file %r", path)
+            encoding = 'utf-8'
+        elif result['encoding'].lower() == 'ascii':
+            # Treat ASCII as UTF-8 (an ASCII document is also valid UTF-8)
+            encoding = 'utf-8'
+        else:
+            encoding = result['encoding'].lower()
+
+        return encoding
+
+
+def iswbound(char):
+    # GPL 2.0 licensed code by Javier Kohen, Sambhav Kothari
+    # from https://github.com/metabrainz/picard-plugins/blob/2.0/plugins/titlecase/titlecase.py
+    """ Checks whether the given character is a word boundary """
+    category = unicodedata.category(char)
+    return 'Zs' == category or 'Sk' == category or 'P' == category[0]
+
+
+def titlecase(text):
+    # GPL 2.0 licensed code by Javier Kohen, Sambhav Kothari
+    # from https://github.com/metabrainz/picard-plugins/blob/2.0/plugins/titlecase/titlecase.py
+    """Converts text to title case following word boundary rules.
+
+    Capitalizes the first character of each word in the input text, where words
+    are determined by Unicode word boundaries. Preserves existing capitalization
+    after the first character of each word.
+
+    Args:
+        text (str): The input text to convert to title case.
+
+    Returns:
+        str: The text converted to title case. Returns empty string if input is empty.
+
+    Examples:
+        >>> titlecase("hello world")
+        'Hello World'
+        >>> titlecase("children's music")
+        'Children's Music'
+    """
+    if not text:
+        return text
+    capitalized = text[0].capitalize()
+    capital = False
+    for i in range(1, len(text)):
+        t = text[i]
+        if t in "’'" and text[i-1].isalpha():
+            capital = False
+        elif iswbound(t):
+            capital = True
+        elif capital and t.isalpha():
+            capital = False
+            t = t.capitalize()
+        else:
+            capital = False
+        capitalized += t
+    return capitalized
